/*-
 * Copyright 2005, 2007 Frank Mayhar.  All Rights Reserved.
 * Borrowed heavily from "ffsrecov," copyright 1999, 2001 John-Mark Gurney.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: inode.c,v 1.3 2007/02/06 05:21:50 frank Exp $
 *
 */

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <timeconv.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <ufs/ufs/ufsmount.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>
#include <libufs.h>

#define NO_IN_EXTERNS
#include "ffs2recov.h"

static uint64_t direct_maxblk;		/* Max block # for a direct block.    */
static uint64_t single_maxblk;		/* Max block # for a single indirect. */
static uint64_t double_maxblk;		/* Max block # for a double indirect. */
static uint64_t triple_maxblk;		/* Max block # for a triple indirect. */
static ufs2_daddr_t *iblk_buf = NULL;
static char *data_buf = NULL;

extern int verbose, debug;

void init_inodes(struct uufsd *);
uint64_t inotodb(struct uufsd *, ino_t);
off_t inotooff(struct uufsd *, ino_t);
int inodetofd(struct uufsd *, struct ufs2_dinode *, int);
int inodetofile(struct uufsd *, struct ufs2_dinode *, const char *);
int getinode(struct uufsd *, void *, struct ufs2_dinode *);
int getinodeblock(struct uufsd *, void *, struct ufs2_dinode *, uint64_t);
void duplicateinode(struct uufsd *, ino_t, const char *);
int isblkininode(struct uufsd *, ufs2_daddr_t, struct ufs2_dinode *);

#define dblksize(fs, dip, lbn) \
	(((lbn) >= NDADDR || (dip)->di_size >= smalllblktosize(fs, (lbn) + 1)) \
		? (uint64_t)(fs)->fs_bsize \
		: (fragroundup(fs, blkoff(fs, (dip)->di_size))))

void
init_inodes(
	struct uufsd *disk)
{
	uint64_t nindir = NINDIR(&disk->d_fs);

	direct_maxblk = NDADDR;
	single_maxblk = (uint64_t)NDADDR + nindir;
	double_maxblk = (uint64_t)NDADDR + nindir * nindir;
	triple_maxblk = (uint64_t)NDADDR + nindir * nindir * nindir;
	iblk_buf = malloc(sizeof(ufs2_daddr_t) * NINDIR(&disk->d_fs));
	data_buf = malloc(disk->d_fs.fs_bsize);
	DPRINTF((stderr, "%s:  direct %llu single %llu double %llu triple %llu\n",
		__FUNCTION__, (unsigned long long)direct_maxblk,
		(unsigned long long)single_maxblk,
		(unsigned long long)double_maxblk,
		(unsigned long long)triple_maxblk));
}

static ufs2_daddr_t
getblknuminode(
	struct uufsd *disk,
	struct ufs2_dinode *inode,
	ufs2_daddr_t blk)
{
	ufs2_daddr_t blktoget = 0;
	ufs2_daddr_t iblk;
	int blko;
	int nindir;

	if ((uint64_t) blk < direct_maxblk)	/* Direct block.                      */
		return(inode->di_db[blk]);
	/*
	 * should convert to a loop 
	 */
	nindir = NINDIR(&disk->d_fs);
					/* Single indirect range.             */
	if ((uint64_t) blk >= direct_maxblk && (uint64_t) blk < single_maxblk) {
		iblk = inode->di_ib[0];	/* Address of indirect block.         */
		bread(disk, fsbtodb(&disk->d_fs, iblk), iblk_buf, disk->d_fs.fs_bsize); /* Get it.  */
		blko = blk - direct_maxblk; /* Offset of entry for block.    */
		blktoget = iblk_buf[blko];
		DPRINTF((stderr, "%s:  SI blk %llu iblk %llu blko %d blktoget %llu\n",
			__FUNCTION__, (unsigned long long)blk,
			(unsigned long long)iblk, blko,
			(unsigned long long)blktoget));
		return(blktoget);
	}
					/* Double indirect range.             */
	if ((uint64_t) blk >= single_maxblk && (uint64_t) blk < double_maxblk) {
		iblk = inode->di_ib[1];	/* Address of first indirect block.   */
		bread(disk, fsbtodb(&disk->d_fs, iblk), iblk_buf, disk->d_fs.fs_bsize); /* Get it.  */
		blko = (blk - direct_maxblk) / nindir;
		DPRINTF((stderr, "%s:  DI, blk %llu iblk %llu blko %d new iblk %llu\n",
			__FUNCTION__, (unsigned long long)blk,
			(unsigned long long)iblk, blko,
			(unsigned long long)iblk_buf[blko]));
		iblk = iblk_buf[blko];	/* Address of single indirect block.  */
		if (iblk == 0)
			return(0);
		bread(disk, fsbtodb(&disk->d_fs, iblk), iblk_buf, disk->d_fs.fs_bsize); /* Get it.  */
		blko = (blk - direct_maxblk) % nindir;
		blktoget = iblk_buf[blko];
		DPRINTF((stderr, "%s:  ... blko %d blktoget %llu\n",
			__FUNCTION__, blko, (unsigned long long)blktoget));
		return(blktoget);
	}
					/* Triple indirect range.             */
	if ((uint64_t) blk >= double_maxblk && (uint64_t) blk < triple_maxblk) {
		iblk = inode->di_ib[2];
		/* First get triple-indirect block.                           */
		fprintf(stderr,
			"can't handle triple indirect blocks.\n");
		blktoget = 0;
	}
	return(blktoget);
}

/*
		blkno = bp->b_un.b_indir2[i];
	if (blkno == 0) {
	    if (level == 0)
		putchar('\n');
	    return 0;
	}
	j = sprintf(tempbuf, "%jd", (intmax_t)blkno);
	if (level == 0) {
	    charssofar += j;
	    if (charssofar >= cpl - 2) {
		putchar('\n');
		charssofar = j;
	    }
	}
	fputs(tempbuf, stdout);
	if (level == 0) {
	    printf(", ");
	    charssofar += 2;
	} else {
	    printf(" =>\n");
	    if (printindir(blkno, level - 1, bufp) == 0)
		return 0;
	}
    }
    if (level == 0)
	putchar('\n');
    return 1;
*/

uint64_t
inotodb(
	struct uufsd *disk,
	ino_t inum)
{ 
        off_t a, b;
        
	a = ino_to_fsbo(&disk->d_fs, inum) * sizeof(struct ufs2_dinode);
	b = fsbtodb(&disk->d_fs, ino_to_fsba(&disk->d_fs, inum));
	return(a + b);
}

off_t
inotooff(
	struct uufsd *disk,
	ino_t inum)
{ 
	off_t a, b, c;
        
	a = ino_to_fsbo(&disk->d_fs, inum) * sizeof(struct ufs2_dinode);
	b = fsbtodb(&disk->d_fs, ino_to_fsba(&disk->d_fs, inum));
	c = b * DEV_BSIZE;
	return(a + c);
}

int
inodetofd(
	struct uufsd *disk,
	struct ufs2_dinode *inode,
	int fd)
{
	uint64_t i;
	ufs2_daddr_t blktoget;
	int st;

	/* If there are no blocks, the data is small enough to fit in the
	 * inode.
	 */
	if (inode->di_blocks == 0) {
		write(fd, (char *)&inode->di_db, inode->di_size);
		return(1);
	}
	for (i = 0; i <= lblkno(&disk->d_fs, inode->di_size); i++) {
		if ((blktoget = getblknuminode(disk, inode, i)) != 0) {
			bread(disk, fsbtodb(&disk->d_fs, blktoget), data_buf, disk->d_fs.fs_bsize);
			lseek(fd, lblktosize(&disk->d_fs, i), SEEK_SET);
			/*
			 * If this is the last block, only write as much as
			 * we need.  Otherwise write a full block.
			 */
			if (i == lblkno(&disk->d_fs, inode->di_size))
				st = write(fd, data_buf, blkoff(&disk->d_fs, inode->di_size));
			else
				st = write(fd, data_buf, disk->d_fs.fs_bsize);
			if (st < 0) {
				perror(__FUNCTION__);
				return(0);
			}
		}
		DPRINTF((stderr, "%s:  %llu blktoget %llu\n", __FUNCTION__,
			(unsigned long long)i, (unsigned long long)blktoget));
	}
	return(1);
}

int
inodetofile(
	struct uufsd *disk,
	struct ufs2_dinode *inode,
	const char *name)
{
	int fd, st;

	if ((fd = open(name, O_CREAT | O_WRONLY | O_TRUNC | O_EXLOCK,
						  inode->di_mode & 07777)) == -1)
		return(-1);
	st = inodetofd(disk, inode, fd);
	close(fd);
	if (!st) {
		if (unlink(name) < 0) {
			perror(__FUNCTION__);
			return(-1);
		}
		fprintf(stderr, "%s:  %s lost\n", __FUNCTION__, name);
	}
	return(0);
}

int
getinode(
	struct uufsd *disk,
	void *data,
	struct ufs2_dinode *inode)
{
	uint64_t i;
	ufs2_daddr_t blktoget;
	int error, dlen;

	/* Sanity check the dinode. */
	if (inode->di_size > (u_int64_t) disk->d_fs.fs_dsize * disk->d_bsize) {
		fprintf(stderr,
			"%s:  inode size (%llu > %llu) too big, skipping.\n",
			__FUNCTION__, (unsigned long long)inode->di_size,
			(unsigned long long)(disk->d_fs.fs_dsize * disk->d_bsize));
		return(0);
	}
	for (i = 0; i <= (inode->di_size >> disk->d_fs.fs_bshift); i++) {
		dlen = lblktosize(&disk->d_fs, i);
		if ((blktoget = getblknuminode(disk, inode, i)) == 0)
			bzero((char *)data + dlen, dblksize(&disk->d_fs, inode, i));
		else
			error = bread(disk, fsbtodb(&disk->d_fs, blktoget),
				  (char *)data + dlen, dblksize(&disk->d_fs, inode, i));
	}
	return(1);
}

int
getinodeblock(
	struct uufsd *disk,
	void *data,
	struct ufs2_dinode *inode,
	uint64_t blk)
{
	ufs2_daddr_t blktoget;
	size_t error, len;

	error = 0;
	len = dblksize(&disk->d_fs, inode, blk);
	if ((blktoget = getblknuminode(disk, inode, blk)) == 0)
		bzero(data, len);
	else
		error = readin(disk, data, blocktooffset(disk, blktoget), len);
	if (error != 0)
		return(-error);
	return(len);
}

#define ino_to_blk(fs, inum)	fsbtodb((fs), ino_to_fsba((fs), (inum)))

void
duplicateinode(
	struct uufsd *disk,
	ino_t inum,
	const char *name)
{
	struct timeval times[2];
	struct direct *dir, *ptr;
	struct ufs2_dinode *inode;
	ufs2_daddr_t blk;
	char *nm, *cp;
	mode_t oldumask;
	int failed, ilen;

	if (verbose)
		fprintf(stderr, "Duping inode %d (%s)\n", inum, name);
	ilen = roundup2(sizeof(struct ufs2_dinode), disk->d_bsize);
	inode = malloc(ilen);
	blk = ino_to_fsba(&disk->d_fs, inum);
	if (bread(disk, blk, inode, ilen) < (ssize_t) sizeof(struct ufs2_dinode)) {
		fprintf(stderr, "short read on inode %d, skipping\n", inum);
		free(inode);
		return;
	}
	failed = 0;
	oldumask = umask(0);
	switch (inode->di_mode & IFMT) {
		case IFREG:
			inodetofile(disk, inode, name);
			break;
		case IFDIR:
			if (mkdir(name, inode->di_mode & 07777) == -1) {
				failed = 1;
				break;
			}
			if (chdir(name) == -1) {
				failed = 1;
				break;
			}
			if ((dir = malloc(blkroundup(&disk->d_fs, inode->di_size))) == NULL) {
				fprintf(stderr,
					"%s:%d memory allocation failed for inode %d, %llu bytes\n",
					__FILE__, __LINE__, inum,
					(unsigned long long)inode->di_size);
				failed = 1;
				break;
			}
			if (!getinode(disk, dir, inode)) {
				failed = 1;
				break;
			}
			for (cp = (char *)dir; cp < (char *)dir + inode->di_size;
							    cp += ptr->d_reclen) {
				ptr = (struct direct *)cp;
				duplicateinode(disk, ptr->d_ino, ptr->d_name);
			}
			free(dir);
			if (chdir("..") == -1) {
				failed = 1;
				break;
			}
			break;
		case IFSOCK:
			fprintf(stderr,
				"don't know how to create sockets (%s)\n",
				name);
			break;
		case IFIFO:
			if (mkfifo(name, inode->di_mode & 07777) == -1)
				failed = 1;
			break;
		case IFLNK:
			if (inode->di_size > 0 &&
				inode->di_size < (u_int64_t) disk->d_fs.fs_maxsymlinklen &&
							 inode->di_blocks == 0) {
				nm = (char *)inode->di_db;
				if (symlink(nm, name) < 0) {
					fprintf(stderr,
						"%s:  symlink %s to %s failed: ",
						__FUNCTION__, name, nm);
					perror("");
				}
			}
			else
				fprintf(stderr, "%s:  Symlink %s lost\n",
							    __FUNCTION__, name);
			break;
		case IFBLK:
		case IFCHR:
			if (mknod(name, inode->di_mode, inode->di_rdev) == -1)
				failed = 1;
			break;
		default:
			failed = 1;
			break;
	}
	if (!failed) {
		times[0].tv_sec = inode->di_atime;
#ifdef DI_NSEC
		times[0].tv_usec = inode->di_atimensec / 1000;
#else
		times[0].tv_usec = 0;
#endif
		times[1].tv_sec = inode->di_mtime;
#ifdef DI_NSEC
		times[1].tv_usec = inode->di_mtimensec / 1000;
#else
		times[1].tv_usec = 0;
#endif
		lutimes(name, times);
		chown(name, inode->di_uid, inode->di_gid);
	}

	umask(oldumask);
	free(inode);
}

int
isblkininode(
	struct uufsd *disk,
	ufs2_daddr_t blk,
	struct ufs2_dinode *inode)
{
	int found;
	static ufs2_daddr_t *blks;
	uint64_t i;

	if (blk == 0)
		return 0;
	if (blks == NULL)
		blks = (ufs2_daddr_t *)malloc(disk->d_bsize);
	found = 0;
	if ((inode->di_mode & IFMT) == IFREG) {
		/*
		 * check all first layer indirect blocks 
		 */
		for (i = 0; !found && i < 3; i++)
			if (blk == inode->di_ib[i])
				found = 1;
		/*
		 * check all second layer indirect blocks 
		 */
		bread(disk, fsbtodb(&disk->d_fs, inode->di_ib[1]), blks, disk->d_bsize);
		for (i = 0; !found && i < (uint64_t) disk->d_fs.fs_nindir; i++)
			if (blk == blks[i])
				found = 1;
		/*
		 * XXX - no support for third layer indirect blocks 
		 */
		for (i = 0; !found && i <= (inode->di_size >> disk->d_fs.fs_bshift); i++)
			if (blk == getblknuminode(disk, inode, i))
				found = 1;
	}

	return(found);
}
