/*
 * Copyright 2005, 2007 Frank Mayhar.  All Rights Reserved.
 * Borrowed heavily from "ffsrecov," copyright 1999, 2001 John-Mark Gurney.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: io.c,v 1.3 2007/02/06 05:21:50 frank Exp $
 *
 */

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <err.h>
#include <timeconv.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <ufs/ufs/ufsmount.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>
#include <libufs.h>

#define NO_IO_EXTERNS
#include "ffs2recov.h"

ssize_t bpread(struct uufsd *, char *, size_t, off_t);
int writeout(struct uufsd *, int, off_t, size_t);
size_t readin(struct uufsd *, void *, off_t, size_t);
void *areadin(struct uufsd *, off_t, size_t);

ssize_t
bpread(
	struct uufsd *disk,
	char *data,
	size_t len,
	off_t off)
{
	off_t rndo;
	size_t rndl;
	size_t t;
	size_t pos;
	size_t olen;
	ufs2_daddr_t blkno;
	static char *buf = NULL;

	if (buf == NULL)
		buf = malloc(disk->d_bsize);
	pos = 0;
	olen = len;
	rndo = off & -disk->d_bsize;

	/* handle odd size first block, or front of block */
	if (off != rndo) {
		blkno = rndo / disk->d_bsize;
		if (bread(disk, fsbtodb(&disk->d_fs, blkno), buf, disk->d_bsize) < 0)
			return(-1);
		if (len < (size_t)(disk->d_bsize - (off - rndo)))
			t = len;
		else
			t = disk->d_bsize - (off - rndo);
		memcpy(data, buf + off - rndo, t);
		pos = t;
		len -= t;
		/* move to the next block */
		rndo += disk->d_bsize;
	}

	/* handle middle block */
	rndl = len & -disk->d_bsize;
	blkno = rndo / disk->d_bsize;
	if (rndl != 0 && bread(disk, fsbtodb(&disk->d_fs, blkno), data + pos, rndl) == -1)
		return(-1);
	pos += rndl;

	/* handle end block */
	if (rndl != len) {
		blkno = (rndo + rndl) / disk->d_bsize;
		if (bread(disk, fsbtodb(&disk->d_fs, blkno), buf, disk->d_bsize) == -1)
			return(-1);
		memcpy(data + pos, buf, len - rndl);
	}
	return(olen);
}

int
writeout(
	struct uufsd *disk,
	int fd,
	off_t pos,
	size_t len)
{
	static int buflen = 0;
	static char *buf = NULL;
	int use_bpread = 0;
	size_t i;
	int j, k, l;

	if ((size_t)buflen < len) {
		if ((buf = realloc(buf, len)) == NULL) {
			perror("malloc");
			exit(1);
		}
		buflen = len;
	}

	i = 0;
	while (i < len) {
		/*
		 * XXX - may return 0, but should never unless fs is
		 * corrupt.
		 */
		if (!use_bpread &&
		    (j = pread(disk->d_fd, buf, len - i, pos + i)) < 0)
			use_bpread = 1;
		if (use_bpread &&
		    (j = bpread(disk, buf, len - i, pos + i)) < 0) {
			fprintf(stderr, "%s:  bpread failed (%llu/%llu)",
				__FUNCTION__, (unsigned long long)(pos + i),
				(unsigned long long)(len - i));
			perror(" ");
			return(0);
		}
		if ((size_t)j < len - i) {
			fprintf(stderr, "%s:  short read (%d/%llu)\n",
				__FUNCTION__, j, (unsigned long long)(len - i));
			return(0);
		}
		k = 0;
		while (k < j) {
			if ((l = write(fd, buf + k, j - k)) == -1) {
				perror("write");
				exit(1);
			}
			k += l;
		}
		i += j;
	}
	return(1);
}

#define min(a, b)	((a) < (b) ? (a) : (b))

size_t
readin(
	struct uufsd *disk,
	void *data,
	off_t pos,
	size_t len)
{
	off_t off, rc;
	size_t olen;
	char *dp;
	ufs2_daddr_t blk;
	static char *buf = NULL;

	if (buf == NULL)
		buf = malloc(disk->d_bsize);
	dp = data;
	blk = pos / disk->d_bsize;
	off = pos % disk->d_bsize;
	olen = len;
	do {
		rc = bread(disk, blk, buf, disk->d_bsize);
		if (rc < 0) {
			fprintf(stderr, "%s:  bread failed (blk %llu len %ld)",
				__FUNCTION__, (unsigned long long)blk,
				disk->d_bsize);
			perror(" ");
			exit(1);
		}
		bcopy(buf + off, dp, min(rc - off, (off_t)len));
		dp += rc - off;
		len -= min(rc - off, (off_t)len);
		off = 0;
		blk++;
	} while (len > 0 && rc == disk->d_bsize);
	return(olen - len);
}

void *
areadin(
	struct uufsd *disk,
	off_t pos,
	size_t len)
{
	void *ret;

	if ((ret = malloc(len)) == NULL) {
		perror("malloc");
		exit(1);
	}
	if (readin(disk, ret, pos, len) != len)
		fprintf(stderr,
		    "WARNING: short read on areadin, data may not be valid.\n");
	return(ret);
}
