/*
 * Copyright 2005, 2007 Frank Mayhar.  All Rights Reserved.
 * Borrowed heavily from "ffsrecov," copyright 1999, 2001 John-Mark Gurney.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: main.c,v 1.3 2007/02/06 05:21:50 frank Exp $
 *
 */

/*
 * This is the published version of this thing.  Bump this when you release
 * a new one.
 */
#define VERSION "1.0"

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <err.h>
#include <ctype.h>
#include <timeconv.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <ufs/ufs/ufsmount.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>
#include <libufs.h>

#include "ffs2recov.h"

struct uufsd *ufs_disk;
struct stat disk_st;

extern int errno;

#define FBLK_ERR	-1
#define FBLK_CG		1
#define FBLK_INODE	2

extern void init_inodes(struct uufsd *);
int verbose;
int debug;

static void
ufserr(const char *name)
{
	if (ufs_disk != NULL && ufs_disk->d_error != NULL)
		warnx("%s: %s", name, ufs_disk->d_error);
	else if (errno != 0)
		warn("%s", name);
}

static void
usage(void)
{
	printf("usage: ffsrecover [-adps] [-b blk] [-c ino -n name | -f ino | -i ino ] [-o blk] [-r name] [-t blk] file\n");
	printf("\n");
	printf("  -a\t\tRecursively recover all inodes in the file system.  Unnamed\n");
	printf("\t\tfiles are named from the inode number.\n");
	printf("  -b <blk>\tFind the inode that owns the indicated block.\n");
	printf("  -c <ino> -n <name> Recursively recover the indicated inode using the given\n");
	printf("\t\tname.\n");
	printf("  -d\t\tPrint all directories and associated inodes.\n");
	printf("  -D\t\tTurn on extra debugging output.\n");
	printf("  -f <ino>\tPrint the contents of the given inode.\n");
	printf("  -i <ino>\tPrint the data for the given inode to stdout.\n");
	printf("  -o <blk>\tUse the given block as the superblock.\n");
	printf("  -p\t\tPrint the parameters of the file system (a la dumpfs).\n");
	printf("  -r <name>\tSearch all directory inodes for the given file and recover it.\n");
	printf("  -S\t\tDon't automatically read the superblock (implied by -o and -s)\n");
	printf("  -s\t\tFind, validate and print all superblocks.\n");
	printf("  -t <blk>\tUse the given block as the first block of the file\n");
	printf("\t\tsystem.\n");
	printf("  -v\t\tIncrease message verbosity level by one.\n");
	printf("  -V\t\tPrint the version and exit.\n");
	exit(1);
}

static void
show_version(void)
{
	printf("ffs2recov version %s.\n", VERSION);
}

int
main(
	int argc,
	char *argv[])
{
	char *file;
	int ch, findsb, printbsb, recoverfile, printalldirs;
	ufs2_daddr_t blk;
	ino_t inum;
	int dumpinode;
	int dumpallfiles;
	int printinfo;
	int createinode;
	int fblk;
	int nosb;
	off_t soff;
	off_t startblk;

	findsb = 0;
	printbsb = 0;
	recoverfile = 0;
	printalldirs = 0;
	dumpinode = 0;
	dumpallfiles = 0;
	printinfo = 0;
	createinode = 0;
	fblk = 0;
	inum = 0;
	file = NULL;
	blk = 0;
	soff = -1;
	startblk = 0;
	nosb = 0;
	verbose = 0;
	debug = 0;

	while ((ch = getopt(argc, argv, "ab:c:dDf:i:n:o:pr:sSt:vV")) != -1)
		switch (ch) {
		case 'a':
			dumpallfiles = 1;
			break;
		case 'b':
			fblk = 1;
			blk = atoi(optarg);
			break;
		case 'c':
			createinode = 1;
			if (inum)
				usage();
			inum = atoi(optarg);
			break;
		case 'd':
			printalldirs = 1;
			break;
		case 'D':
			debug = 1;
			break;
		case 'f':
			printinfo = 1;
			if (inum)
				usage();
			inum = atoi(optarg);
			break;
		case 'i':
			dumpinode = 1;
			if (inum)
				usage();
			inum = atoi(optarg);
			break;
		case 'n':
			if (file != NULL)
				usage();
			file = strdup(optarg);
			break;
		case 'o':
			soff = strtoq(optarg, NULL, 0);
			nosb = 1;
			break;
		case 'p':
			printbsb = 1;
			break;
		case 'r':
			recoverfile = 1;
			file = strdup(optarg);
			break;
		case 's':
			findsb = 1;
			nosb = 1;
			break;
		case 'S':
			nosb = 1;
			break;
		case 't':
			startblk = strtoq(optarg, NULL, 0);
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			show_version();
			exit(0);
			/* NOTREACHED */
		case '?':
		default:
			usage();
		}

	argc -= optind;
	argv += optind;

	if (argc == 0 || argc > 1)
		usage();

	if (createinode && file == NULL)
		usage();

	if ((ufs_disk = malloc(sizeof(struct uufsd))) == NULL) {
		fprintf(stderr, "No memory for struct ufs_disk\n");
		exit(2);
	}
	if (ufs_disk_fillout_blank(ufs_disk, *argv) == -1) {
		ufserr(*argv);
		exit(1);
	}
	if (fstat(ufs_disk->d_fd, &disk_st) == -1) {
		perror("stat");
		exit(1);
	}
	if (nosb == 0 && sbread(ufs_disk) < 0) {
		ufserr(*argv);
		exit(1);
	}
	else
		ufs_disk->d_bsize = disk_st.st_blksize;
	if (soff > 0) {
		int rc;

		rc = get_sblock(ufs_disk, soff);
		if (rc < 0)
			exit(2);
	}
	init_inodes(ufs_disk);
/*
	if ((t = validatesb(&ufs_disk)) != NULL)
		fprintf(stderr,
		   "WARNING: field %s is invalid, superblock may be corrupt.\n",
		    t);

	if (ufs_disk->fs.fs_magic != FS_MAGIC)
		fprintf(stderr, "doesn't have magic\n");
*/
	if (createinode)
		duplicateinode(ufs_disk, inum, file);

	if (dumpallfiles)
		dumpall(ufs_disk);

	if (dumpinode)
		dumpoutinode(ufs_disk, inum);

	if (printalldirs)
		printdirs(ufs_disk);

	if (findsb)
		findsuperblks(ufs_disk);

	if (printinfo)
		printinodeinfo(ufs_disk, inum);

	if (printbsb)
		dumpfs(ufs_disk);

	if (recoverfile)
		searchfsforname(ufs_disk, file);

	if (fblk)
		findblk(ufs_disk, blk, &inum);

	return 0;
}


void
printdirect(
	struct direct *dir,
	int size)
{
	struct direct *ptr;
	char *c, *np;

	if (dir == NULL || size == 0)
		return;
	for (c = (char *)dir; c < (char *)dir + size; c += ptr->d_reclen) {
		ptr = (struct direct *)c;
		for (np = ptr->d_name; *np != '\0'; np++)
			if (!isgraph(*np) || *np == '/')
				break;
		if (*np == '\0')
			printf("(%d, \"%s\"), ", ptr->d_ino, ptr->d_name);
		if (ptr->d_reclen <= 0)
			break;
	}
}

void
printdirs(
	struct uufsd *disk)
{
	struct ufs2_dinode inode;
	struct direct *dir;
	int i;

	for (i = 2; i < disk->d_fs.fs_ncg * disk->d_fs.fs_ipg; i++) {
		if (verbose)
			fprintf(stderr, "Inode %d\r", i);
		if (readin(disk, &inode, inotooff(disk, i), sizeof(inode)) != sizeof(inode)) {
			if (verbose > 1)
				fprintf(stderr, "\nskip short\n");
			continue;
		}
		if (i == 2 || (inode.di_mode & IFMT) == IFDIR) {
			/*
			 * Sanity check the dinode.  The directory can't
			 * possibly be bigger than the file system itself.
			 */
			if (inode.di_size > (uint64_t) disk->d_fs.fs_dsize * disk->d_bsize) {
				if (verbose > 1)
					fprintf(stderr, "\ninode too big\n");
				continue;
			}
			/*
			 * If a directory is bigger than a gigabyte it's
			 * probably too big, but we want to complain to the
			 * user, just in case.
			 */
			if (inode.di_size > 1024 * 1024 * 1024) {
				fprintf(stderr,
					"%s:  Directory inode %d bigger than 1GiB (%qd), skipping.\n",
					__FUNCTION__, i, (unsigned long long)inode.di_size);
				continue;
			}
			if ((dir = malloc(blkroundup(&disk->d_fs, inode.di_size))) == NULL) {
				if (verbose)
					fprintf(stderr, "\n");
				fprintf(stderr,
				    "%s:  memory allocation failed for inode %d, %llu bytes\n",
				    __FUNCTION__, i, (unsigned long long)inode.di_size);
				continue;
			}
			if (getinode(disk, dir, &inode)) {
				printf("inode %d: ", i);
				printdirect(dir, inode.di_size);
				printf("\n");
			}
			free(dir);
		}
	}
}

void
searchfsforname(
	struct uufsd *disk,
	char *name)
{
	struct ufs2_dinode *inode;
	ufs2_daddr_t blk;
	struct direct *dir;
	int i, j, len, ilen;
	ino_t *inol = 0;

	ilen = roundup2(sizeof(struct ufs2_dinode), disk->d_bsize);
	inode = malloc(ilen);
	for (i = 2; i < disk->d_fs.fs_ncg * disk->d_fs.fs_ipg; i++) {
		blk = ino_to_fsba(&disk->d_fs, i);
		if (verbose)
			fprintf(stderr, "Inode %d blk %llu\r", i, (unsigned long long)blk);
		if ((len = bread(disk, blk, inode, ilen)) < (ssize_t) sizeof(struct ufs2_dinode)) {
			if (verbose > 1)
				fprintf(stderr,
					"\n%s:  short read (%d) on inode %d blk %llu, skipping\n",
						     __FUNCTION__, len, i, (unsigned long long)blk);
			continue;
		}
		if (i == 2 || (inode->di_mode & IFMT) == IFDIR) {
			len = (inode->di_size + disk->d_fs.fs_fsize - 1) &
							    disk->d_fs.fs_fmask;
			if ((dir = malloc(len)) == NULL) {
				if (verbose)
					fprintf(stderr, "\n");
				fprintf(stderr,
				    "%s:%d memory allocation failed for inode %d, %d bytes\n",
				    __FILE__, __LINE__, i, len);
				continue;
			}
			if (getinode(disk, dir, inode)) {
				inol = matchname(disk, inode, name);
				for (j = 0; inol[j] != 0; j++)
					printf("matched: %6d\tparent: %6d\n",
								    inol[j], i);
			}
			free(inol);
			free(dir);
		}
	}
	free(inode);
}

ino_t *
matchname(
	struct uufsd *disk,
	struct ufs2_dinode *dir,
	char *name)
{
	ino_t *list;
	ino_t *nlist;
	int lalloc;
	void *data;
	char *c;
	struct direct *ptr;
	int len;

	len = (dir->di_size + disk->d_fs.fs_fsize - 1) & disk->d_fs.fs_fmask;
	if ((data = malloc(len)) == NULL) {
		fprintf(stderr, "%s:%d memory allocation failed for directory, %d bytes\n",
			__FILE__, __LINE__, len);
		return(NULL);
	}
	if ((list = malloc(sizeof *list * (lalloc = 1))) == NULL) {
		fprintf(stderr, "memory allocation failed for directory\n");
		return(NULL);
	}

	if (!getinode(disk, data, dir)) {
		free(data);
		free(list);
		return(NULL);
	}

	list[lalloc - 1] = 0;
	for (c = data; c < (char *) data + dir->di_size; c += ptr->d_reclen) {
		ptr = (struct direct *)c;
		if (strcmp(ptr->d_name, name) == 0) {
			if ((nlist = realloc(list, ++lalloc)) == NULL) {
				fprintf(stderr,
				    "can't alocate more memory, aborting\n");
				free(data);
				free(list);
				return NULL;
			}
			list = nlist;
			list[lalloc - 2] = ptr->d_ino;
			list[lalloc - 1] = 0;
		}
	}
	free(data);

	return(list);
}

/*
 * XXX - for large files this is not optimal, should make inode block
 * traversal a function, and then interate over it here
 */
void
dumpoutinode(
	struct uufsd *disk,
	ino_t num)
{
	struct ufs2_dinode *inode;
	ufs2_daddr_t blk;
	int st, ilen;

	ilen = roundup2(sizeof(struct ufs2_dinode), disk->d_bsize);
	inode = malloc(ilen);
	blk = ino_to_fsba(&disk->d_fs, num);
	DPRINTF((stderr, "%s:  ino %d blk %llu\n", __FUNCTION__, num, (unsigned long long)blk));
	if (bread(disk, blk, inode, ilen) < (ssize_t) sizeof(struct ufs2_dinode)) {
		fprintf(stderr, "short read on inode %d, skipping\n", num);
		free(inode);
		return;
	}
	st = inodetofd(disk, inode, STDOUT_FILENO);
	free(inode);
}

void
dumpall(
	struct uufsd *disk)
{
	struct ufs2_dinode *inode;
	ufs2_daddr_t blk;
	void *data;
	int i, ilen;
	char name[16];

	ilen = roundup2(sizeof(struct ufs2_dinode), disk->d_bsize);
	inode = malloc(ilen);
	for (i = 2; i < disk->d_fs.fs_ncg * disk->d_fs.fs_ipg; i++) {
		blk = ino_to_fsba(&disk->d_fs, i);
		if (verbose)
			fprintf(stderr, "Inode %d blk %llu\r", i, (unsigned long long)blk);
		if (bread(disk, blk, inode, ilen) < (ssize_t) sizeof(struct ufs2_dinode)) {
			fprintf(stderr, "short read on inode %d, skipping\n", i);
			continue;
		}
		if ((inode->di_mode & IFMT) == IFREG) {
			int len;

			len = (inode->di_size + disk->d_fs.fs_fsize - 1) &
							    disk->d_fs.fs_fmask;
			if ((data = malloc(len)) == NULL) {
				if (verbose)
					fprintf(stderr, "\n");
				fprintf(stderr,
				    "%s:%d memory allocation failed for inode %d, %d bytes\n",
				    __FILE__, __LINE__, i, len);
				continue;
			}
			if (getinode(disk, data, inode)) {
				snprintf(name, sizeof name, "%d", i);
				inodetofile(disk, inode, name);
			}
			free(data);
		}
	}
	free(inode);
}

void
printinodeinfo(
	struct uufsd *disk,
	ino_t inum)
{
	struct ufs2_dinode *inode;
	ufs2_daddr_t blk;
	int ilen;

	ilen = roundup2(sizeof(struct ufs2_dinode), disk->d_bsize);
	inode = malloc(ilen);
	blk = ino_to_fsba(&disk->d_fs, inum);
	if (bread(disk, blk, inode, ilen) < (ssize_t) sizeof(struct ufs2_dinode)) {
		fprintf(stderr, "short read on inode %d, skipping\n", inum);
		free(inode);
		return;
	}

	printf("inum: %d\n", inum);
	printdinode(inode);
	free(inode);
}

void
printdinode(
	struct ufs2_dinode *inode)
{
	time_t ct;

	printf("mode: %s\n", typestr(inode->di_mode & IFMT));
	printf("perm: %o\n", inode->di_mode);
	printf("link: %d\n", inode->di_nlink);
	printf("size: %qd\n", (unsigned long long)inode->di_size);
	ct = _time64_to_time(inode->di_mtime);
	printf("mtim: %s", ctime(&ct));
	ct = _time64_to_time(inode->di_ctime);
	printf("ctim: %s", ctime(&ct));
	ct = _time64_to_time(inode->di_atime);
	printf("atim: %s", ctime(&ct));
	printf("flgs: %d\n", inode->di_flags);
	printf("blks: %llu\n", (unsigned long long)inode->di_blocks);
	printf(" gen: 0x%x\n", inode->di_gen);
	printf(" uid: %d\n", inode->di_uid);
	printf(" gid: %d\n", inode->di_gid);
	printf("rdev: 0x%x\n", (unsigned int)inode->di_rdev);
#ifdef DI_SPARE
	printf("spr1: 0x%x\n", (unsigned int)inode->di_spare[0]);
	printf("spr2: 0x%x\n", (unsigned int)inode->di_spare[1]);
#endif
}

const char *
typestr(
	int type)
{
	static char buf[32];
	if (type == IFIFO)
		return "pipe";
	if (type == IFCHR)
		return "character";
	if (type == IFDIR)
		return "directory";
	if (type == IFBLK)
		return "block";
	if (type == IFREG)
		return "regular";
	if (type == IFLNK)
		return "symlink";
	if (type == IFSOCK)
		return "socket";
	snprintf(buf, sizeof buf, "invalid type 0x%x", type);
	return buf;
}
