/*
 * Copyright 2005, 2007 Frank Mayhar.  All Rights Reserved.
 * Borrowed heavily from "ffsrecov," copyright 1999, 2001 John-Mark Gurney.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: super.c,v 1.3 2007/02/06 05:21:50 frank Exp $
 *
 */

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <inttypes.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <ufs/ufs/ufsmount.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>
#include <libufs.h>

#define	acg	disk.d_cg

#define NO_SB_EXTERNS
#include "ffs2recov.h"

extern struct stat disk_st;
extern int verbose;

void checksbcg(struct uufsd *, off_t);
void findsuperblks(struct uufsd *);
struct cg *get_cg(struct uufsd *, struct fs *, off_t);
int get_sblock(struct uufsd *, off_t);
void pbits(void *, int);
int dumpcg(struct uufsd *);
int dumpfs(struct uufsd *);

void
checksbcg(
	struct uufsd *disk,
	off_t pos)
{
	struct fs *fs;
	struct cg *cg;
	off_t cgpos;
	off_t fsb;
	off_t sboff;

	fs = areadin(disk, pos, SBLOCKSIZE);
	if (fs->fs_magic != FS_UFS2_MAGIC)
		goto nocg;

	printf("superblock: %llu(byte), %llu(block)",
		(unsigned long long)pos, (unsigned long long)(pos / DEV_BSIZE));

	/*
	 * offset of block plus offset of known cyl group minus offset of known
	 * superblock
	 */
	sboff = fsbtodb(fs, fs->fs_sblkno) * DEV_BSIZE;
	cgpos = pos + fsbtodb(fs, fs->fs_cblkno) * DEV_BSIZE - sboff;
	/*
	 * this is to prevent outragious struct fs's from causing problems,
	 * no cg should be more than 64k, though this can still cause FP
	 * issues.
	 */
	cg = NULL;
	if (CGSIZE(fs) < 64 * 1024) {
		if (verbose > 1)
			fprintf(stderr,
				"Reading CG @%llu (pos %llu sboff %llu cblkno %d sblkno %d)\n",
				(unsigned long long)cgpos,
				(unsigned long long)pos,
				(unsigned long long)sboff, fs->fs_cblkno,
				fs->fs_sblkno);
		cg = areadin(disk, cgpos, CGSIZE(fs));
		if (cg->cg_magic == CG_MAGIC) {
			printf(", cg: %llu(byte), %llu(block), %llu(nth)",
				(unsigned long long)cgpos,
				(unsigned long long)(cgpos / DEV_BSIZE),
				(unsigned long long)cg->cg_cgx);
			fsb = (pos - sboff) / DEV_BSIZE - fsbtodb(fs,
							cgbase(fs, cg->cg_cgx));
			printf(", fs begin: %llu(block), primary sb at: %llu(block)",
				(unsigned long long)fsb,
				(unsigned long long)(fsb + SBLOCK_UFS2 / DEV_BSIZE));
		}
	}

	printf("\n");

	free(cg);
nocg:
	free(fs);
}

void
findsuperblks(
	struct uufsd *disk)
{
	off_t *sblocks;
	int32_t *sbmagic;
	int sblockscnt;
	uint64_t i;
	off_t sbpos;
	int moff;
	int mdif;
	int32_t *tm;
	off_t end;
	size_t sr;

	sblocks = NULL;
	sblockscnt = 0;

	if ((sbmagic = malloc(INT_BLKSIZE)) == NULL) {
		perror("malloc");
		exit(1);
	}

	moff = ((int)(&((struct fs *)NULL)->fs_magic)) % DEV_BSIZE;
	mdif = ((int)(&((struct fs *)NULL)->fs_magic)) & -DEV_BSIZE;

	if (disk_st.st_size == 0 || (disk_st.st_mode & S_IFMT) == S_IFCHR)
		end = (off_t)LLONG_MAX;
	else
		end = disk_st.st_size;

	for (i = 0; i < (uint64_t)end; i += DEV_BSIZE) {
		if (verbose)
			fprintf(stderr, "Offset %llu\r", (unsigned long long)i);
		if ((i % INT_BLKSIZE) == 0)
			if ((sr = readin(disk, sbmagic, i, INT_BLKSIZE)) != INT_BLKSIZE)
				end = i + sr;
		tm = (int32_t *)((char *)sbmagic + i % INT_BLKSIZE + moff);
		if (*(int32_t *)((char *)sbmagic + i % INT_BLKSIZE + moff) == FS_UFS2_MAGIC) {
			sbpos = i - mdif;
			if ((sblocks = realloc(sblocks, ++sblockscnt * sizeof(*sblocks))) == NULL) {
				perror("realloc");
				exit(1);
			}
			sblocks[sblockscnt - 1] = sbpos;
			checksbcg(disk, sbpos);
			fflush(stdout);
		}
	}
	free(sbmagic);
	free(sblocks);
}

static const char *
validatesb(
	struct fs *fs)
{
	if (fs->fs_bmask != ~(fs->fs_bsize - 1))
		return "fs_bmask";
	if (fs->fs_fmask != ~(fs->fs_fsize - 1))
		return "fs_fmask";
	if ((1 << fs->fs_bshift) != fs->fs_bsize)
		return "fs_bshift";
	if ((1 << fs->fs_fshift) != fs->fs_fsize)
		return "fs_fshift";
	if (fs->fs_fragshift != fs->fs_bshift - fs->fs_fshift)
		return "fs_fragshift";
	if ((DEV_BSIZE << fs->fs_fsbtodb) != fs->fs_fsize)
		return "fs_fsbtodb";
	/* need to check sbsize, csmask, csshift */
	if (fs->fs_nindir != fs->fs_bsize / (int32_t)sizeof(ufs2_daddr_t))
		return "fs_nindir";
	if (fs->fs_inopb != fs->fs_bsize / (int32_t)sizeof(struct ufs2_dinode))
		return "fs_inopb";
	/* need to check csaddr, cssize, cgsize */
	/* need to check cpg, ipg, fpg */
#ifdef REAL_QUAD_T
	if (fs->fs_qbmask != fs->fs_bsize - 1)
		return "fs_qbmask";
	if (fs->fs_qfmask != fs->fs_fsize - 1)
		return "fs_qfmask";
#else
	if (*(int64_t *)&fs->fs_qbmask != fs->fs_bsize - 1)
		return "fs_qbmask";
	if (*(int64_t *)&fs->fs_qfmask != fs->fs_fsize - 1)
		return "fs_qfmask";
#endif

	return NULL;
}

struct cg *
get_cg(
	struct uufsd *disk,
	struct fs *fs,
	off_t pos)
{
	struct cg *cg;
	off_t cgpos;
	off_t sboff;

	cg = NULL;
	if (fs->fs_magic != FS_UFS2_MAGIC)
		goto nocg;
	/*
	 * offset of block plus offset of known cyl group minus offset of known
	 * superblock
	 */
	sboff = fsbtodb(fs, fs->fs_sblkno) * DEV_BSIZE;
	cgpos = pos + fsbtodb(fs, fs->fs_cblkno) * DEV_BSIZE - sboff;
	/*
	 * this is to prevent outragious struct fs's from causing problems,
	 * no cg should be more than 64k, though this can still cause FP
	 * issues.
	 */
	if (CGSIZE(fs) < 64 * 1024) {
		if (verbose > 1)
			fprintf(stderr,
				"Reading CG @%llu (pos %llu sboff %llu cblkno %d sblkno %d)\n",
				(unsigned long long)cgpos,
				(unsigned long long)pos,
				(unsigned long long)sboff, fs->fs_cblkno,
				fs->fs_sblkno);
		cg = areadin(disk, cgpos, CGSIZE(fs));
		if (cg->cg_magic != CG_MAGIC) {
			free(cg);
			cg = NULL;
		}
	}
nocg:	;
	return(cg);
}

int
get_sblock(
	struct uufsd *disk,
	off_t soff)
{
	size_t rc;
	struct fs *fs;
	struct cg *cg;
	const char *fld;

	fs = malloc(SBLOCKSIZE);
	rc = readin(disk, fs, soff, SBLOCKSIZE);
	if (rc < SBLOCKSIZE) {
		fprintf(stderr, "%s:  Short read for superblock\n", __FUNCTION__);
		free(fs);
		return(-1);
	}
	if (fs->fs_magic != FS_UFS2_MAGIC) {
		fprintf(stderr, "%s:  Bad magic (%x)\n", __FUNCTION__, fs->fs_magic);
		free(fs);
		return(-1);
	}
	if ((fld = validatesb(fs)) != NULL)
		fprintf(stderr, "%s:  Field %s bad.\n", __FUNCTION__, fld);
	bcopy(fs, &disk->d_fs, sizeof(struct fs));
	disk->d_ufs = 2;
	cg = get_cg(disk, fs, soff);
	if (cg != NULL) {
		bcopy(cg, &disk->d_cg, sizeof(struct cg));
		free(cg);
	}
	free(fs);
	return(0);
}

void
pbits(void *vp, int max)
{
	int i;
	char *p;
	int count, j;

	for (count = i = 0, p = vp; i < max; i++)
		if (isset(p, i)) {
			if (count)
				printf(",%s", count % 6 ? " " : "\n\t");
			count++;
			printf("%d", i);
			j = i;
			while ((i+1)<max && isset(p, i+1))
				i++;
			if (i != j)
				printf("-%d", i);
		}
	printf("\n");
}

int
dumpcg(
	struct uufsd *disk)
{
	time_t cgtime;
	off_t cur;
	int i, j;

	printf("\ncg %d:\n", disk->d_lcg);
	cur = fsbtodb(&disk->d_fs, cgtod(&disk->d_fs, disk->d_lcg)) * disk->d_bsize;
	switch (disk->d_ufs) {
	case 2:
		cgtime = disk->d_cg.cg_time;
		printf("magic\t%x\ttell\t%jx\ttime\t%s",
		    disk->d_cg.cg_magic, (intmax_t)cur, ctime(&cgtime));
		printf("cgx\t%d\tndblk\t%d\tniblk\t%d\tinitiblk %d\n",
		    disk->d_cg.cg_cgx, disk->d_cg.cg_ndblk, disk->d_cg.cg_niblk, disk->d_cg.cg_initediblk);
		break;
	case 1:
		cgtime = disk->d_cg.cg_old_time;
		printf("magic\t%x\ttell\t%jx\ttime\t%s",
		    disk->d_cg.cg_magic, (intmax_t)cur, ctime(&cgtime));
		printf("cgx\t%d\tncyl\t%d\tniblk\t%d\tndblk\t%d\n",
		    disk->d_cg.cg_cgx, disk->d_cg.cg_old_ncyl, disk->d_cg.cg_old_niblk,
		    disk->d_cg.cg_ndblk);
		break;
	default:
		break;
	}
	printf("nbfree\t%d\tndir\t%d\tnifree\t%d\tnffree\t%d\n",
	    disk->d_cg.cg_cs.cs_nbfree, disk->d_cg.cg_cs.cs_ndir,
	    disk->d_cg.cg_cs.cs_nifree, disk->d_cg.cg_cs.cs_nffree);
	printf("rotor\t%d\tirotor\t%d\tfrotor\t%d\nfrsum",
	    disk->d_cg.cg_rotor, disk->d_cg.cg_irotor, disk->d_cg.cg_frotor);
	for (i = 1, j = 0; i < disk->d_fs.fs_frag; i++) {
		printf("\t%d", disk->d_cg.cg_frsum[i]);
		j += i * disk->d_cg.cg_frsum[i];
	}
	printf("\nsum of frsum: %d", j);
	if (disk->d_fs.fs_contigsumsize > 0) {
		for (i = 1; i < disk->d_fs.fs_contigsumsize; i++) {
			if ((i - 1) % 8 == 0)
				printf("\nclusters %d-%d:", i,
				    disk->d_fs.fs_contigsumsize - 1 < i + 7 ?
				    disk->d_fs.fs_contigsumsize - 1 : i + 7);
			printf("\t%d", cg_clustersum(&disk->d_cg)[i]);
		}
		printf("\nclusters size %d and over: %d\n",
		    disk->d_fs.fs_contigsumsize,
		    cg_clustersum(&disk->d_cg)[disk->d_fs.fs_contigsumsize]);
		printf("clusters free:\t");
		pbits(cg_clustersfree(&disk->d_cg), disk->d_cg.cg_nclusterblks);
	} else
		printf("\n");
	printf("inodes used:\t");
	pbits(cg_inosused(&disk->d_cg), disk->d_fs.fs_ipg);
	printf("blks free:\t");
	pbits(cg_blksfree(&disk->d_cg), disk->d_fs.fs_fpg);
	return (0);
}

int
dumpfs(
	struct uufsd *disk)
{
	time_t fstime;
	int64_t fssize;
	int32_t fsflags;
	int i;

	switch (disk->d_ufs) {
		case 2:
			fssize = disk->d_fs.fs_size;
			fstime = disk->d_fs.fs_time;
			printf("magic\t%x (UFS2)\ttime\t%s",
						  disk->d_fs.fs_magic, ctime(&fstime));
			printf("superblock location\t%jd\tid\t[ %x %x ]\n",
					(intmax_t)disk->d_fs.fs_sblockloc,
						    disk->d_fs.fs_id[0], disk->d_fs.fs_id[1]);
			printf("ncg\t%d\tsize\t%jd\tblocks\t%jd\n",
					disk->d_fs.fs_ncg, (intmax_t)fssize,
							(intmax_t)disk->d_fs.fs_dsize);
			break;
		case 1:
			fssize = disk->d_fs.fs_old_size;
			fstime = disk->d_fs.fs_old_time;
			printf("magic\t%x (UFS1)\ttime\t%s",
						  disk->d_fs.fs_magic, ctime(&fstime));
			printf("id\t[ %x %x ]\n", disk->d_fs.fs_id[0], disk->d_fs.fs_id[1]);
			printf("ncg\t%d\tsize\t%jd\tblocks\t%jd\n",
					disk->d_fs.fs_ncg, (intmax_t)fssize,
							(intmax_t)disk->d_fs.fs_dsize);
			break;
		default:
			goto err;
	}
	printf("bsize\t%d\tshift\t%d\tmask\t0x%08x\n",
				     disk->d_fs.fs_bsize, disk->d_fs.fs_bshift, disk->d_fs.fs_bmask);
	printf("fsize\t%d\tshift\t%d\tmask\t0x%08x\n",
				     disk->d_fs.fs_fsize, disk->d_fs.fs_fshift, disk->d_fs.fs_fmask);
	printf("frag\t%d\tshift\t%d\tfsbtodb\t%d\n",
				 disk->d_fs.fs_frag, disk->d_fs.fs_fragshift, disk->d_fs.fs_fsbtodb);
	printf("minfree\t%d%%\toptim\t%s\tsymlinklen %d\n",
		disk->d_fs.fs_minfree, disk->d_fs.fs_optim == FS_OPTSPACE ? "space" : "time",
							  disk->d_fs.fs_maxsymlinklen);
	switch (disk->d_ufs) {
		case 2:
			printf("%s %d\tmaxbpg\t%d\tmaxcontig %d\tcontigsumsize %d\n",
			    "maxbsize", disk->d_fs.fs_maxbsize, disk->d_fs.fs_maxbpg,
			    disk->d_fs.fs_maxcontig, disk->d_fs.fs_contigsumsize);
			printf("nbfree\t%jd\tndir\t%jd\tnifree\t%jd\tnffree\t%jd\n",
			    (intmax_t)disk->d_fs.fs_cstotal.cs_nbfree, 
			    (intmax_t)disk->d_fs.fs_cstotal.cs_ndir,
			    (intmax_t)disk->d_fs.fs_cstotal.cs_nifree, 
			    (intmax_t)disk->d_fs.fs_cstotal.cs_nffree);
			printf("bpg\t%d\tfpg\t%d\tipg\t%d\n",
			    disk->d_fs.fs_fpg / disk->d_fs.fs_frag, disk->d_fs.fs_fpg, disk->d_fs.fs_ipg);
			printf("nindir\t%d\tinopb\t%d\tmaxfilesize\t%ju\n",
			    disk->d_fs.fs_nindir, disk->d_fs.fs_inopb, 
			    (uintmax_t)disk->d_fs.fs_maxfilesize);
			printf("sbsize\t%d\tcgsize\t%d\tcsaddr\t%jd\tcssize\t%d\n",
			    disk->d_fs.fs_sbsize, disk->d_fs.fs_cgsize, (intmax_t)disk->d_fs.fs_csaddr,
			    disk->d_fs.fs_cssize);
			break;
		case 1:
			printf("maxbpg\t%d\tmaxcontig %d\tcontigsumsize %d\n",
			    disk->d_fs.fs_maxbpg, disk->d_fs.fs_maxcontig, disk->d_fs.fs_contigsumsize);
			printf("nbfree\t%d\tndir\t%d\tnifree\t%d\tnffree\t%d\n",
			    disk->d_fs.fs_old_cstotal.cs_nbfree, disk->d_fs.fs_old_cstotal.cs_ndir,
			    disk->d_fs.fs_old_cstotal.cs_nifree, disk->d_fs.fs_old_cstotal.cs_nffree);
			printf("cpg\t%d\tbpg\t%d\tfpg\t%d\tipg\t%d\n",
			    disk->d_fs.fs_old_cpg, disk->d_fs.fs_fpg / disk->d_fs.fs_frag, disk->d_fs.fs_fpg,
			    disk->d_fs.fs_ipg);
			printf("nindir\t%d\tinopb\t%d\tnspf\t%d\tmaxfilesize\t%ju\n",
			    disk->d_fs.fs_nindir, disk->d_fs.fs_inopb, disk->d_fs.fs_old_nspf,
			    (uintmax_t)disk->d_fs.fs_maxfilesize);
			printf("sbsize\t%d\tcgsize\t%d\tcgoffset %d\tcgmask\t0x%08x\n",
			    disk->d_fs.fs_sbsize, disk->d_fs.fs_cgsize, disk->d_fs.fs_old_cgoffset,
			    disk->d_fs.fs_old_cgmask);
			printf("csaddr\t%d\tcssize\t%d\n",
			    disk->d_fs.fs_old_csaddr, disk->d_fs.fs_cssize);
			printf("rotdelay %dms\trps\t%d\ttrackskew %d\tinterleave %d\n",
			    disk->d_fs.fs_old_rotdelay, disk->d_fs.fs_old_rps, disk->d_fs.fs_old_trackskew,
			    disk->d_fs.fs_old_interleave);
			printf("nsect\t%d\tnpsect\t%d\tspc\t%d\n",
			    disk->d_fs.fs_old_nsect, disk->d_fs.fs_old_npsect, disk->d_fs.fs_old_spc);
			break;
		default:
			goto err;
	}
	printf("sblkno\t%d\tcblkno\t%d\tiblkno\t%d\tdblkno\t%d\n",
	    disk->d_fs.fs_sblkno, disk->d_fs.fs_cblkno, disk->d_fs.fs_iblkno, disk->d_fs.fs_dblkno);
	printf("cgrotor\t%d\tfmod\t%d\tronly\t%d\tclean\t%d\n",
	    disk->d_fs.fs_cgrotor, disk->d_fs.fs_fmod, disk->d_fs.fs_ronly, disk->d_fs.fs_clean);
	printf("avgfpdir %d\tavgfilesize %d\n",
	    disk->d_fs.fs_avgfpdir, disk->d_fs.fs_avgfilesize);
	printf("flags\t");
	if (disk->d_fs.fs_old_flags & FS_FLAGS_UPDATED)
		fsflags = disk->d_fs.fs_flags;
	else
		fsflags = disk->d_fs.fs_old_flags;
	if (fsflags == 0)
		printf("none");
	if (fsflags & FS_UNCLEAN)
		printf("unclean ");
	if (fsflags & FS_DOSOFTDEP)
		printf("soft-updates ");
	if (fsflags & FS_NEEDSFSCK)
		printf("needs fsck run ");
	if (fsflags & FS_INDEXDIRS)
		printf("indexed directories ");
	if (fsflags & FS_ACLS)
		printf("acls ");
	if (fsflags & FS_MULTILABEL)
		printf("multilabel ");
	if (fsflags & FS_FLAGS_UPDATED)
		printf("fs_flags expanded ");
	fsflags &= ~(FS_UNCLEAN | FS_DOSOFTDEP | FS_NEEDSFSCK | FS_INDEXDIRS |
		     FS_ACLS | FS_MULTILABEL | FS_FLAGS_UPDATED);
	if (fsflags != 0)
		printf("unknown flags (%#x)", fsflags);
	putchar('\n');
	printf("fsmnt\t%s\n", disk->d_fs.fs_fsmnt);
	printf("volname\t%s\tswuid\t%ju\n",
		disk->d_fs.fs_volname, (uintmax_t)disk->d_fs.fs_swuid);
	printf("\ncs[].cs_(nbfree,ndir,nifree,nffree):\n\t");
	disk->d_fs.fs_csp = calloc(1, disk->d_fs.fs_cssize);
	if (bread(disk, fsbtodb(&disk->d_fs, disk->d_fs.fs_csaddr), disk->d_fs.fs_csp, disk->d_fs.fs_cssize) == -1)
		goto err;
	for (i = 0; i < disk->d_fs.fs_ncg; i++) {
		struct csum *cs = &disk->d_fs.fs_cs(&disk->d_fs, i);
		if (i && i % 4 == 0)
			printf("\n\t");
		printf("(%d,%d,%d,%d) ",
		    cs->cs_nbfree, cs->cs_ndir, cs->cs_nifree, cs->cs_nffree);
	}
	printf("\n");
	if (fssize % disk->d_fs.fs_fpg) {
		if (disk->d_ufs == 1)
			printf("cylinders in last group %d\n",
			    howmany(disk->d_fs.fs_old_size % disk->d_fs.fs_fpg,
			    disk->d_fs.fs_old_spc / disk->d_fs.fs_old_nspf));
		printf("blocks in last group %ld\n\n",
		    (long)((fssize % disk->d_fs.fs_fpg) / disk->d_fs.fs_frag));
	}
	while ((i = cgread(disk)) != 0) {
		if (i == -1 || dumpcg(disk))
			goto err;
	}
	return (0);

err:	printf("dumpfs failed\n");
	return (1);
}
